from typing import TYPE_CHECKING

from jsonschema_spec.handlers.file import FileHandler
from jsonschema_spec.handlers.urllib import UrllibHandler

if TYPE_CHECKING:
    from jsonschema_spec.handlers.urllib import UrllibHandler as UrlHandler
else:
    try:
        from jsonschema_spec.handlers.requests import (
            UrlRequestsHandler as UrlHandler,
        )
    except ImportError:
        from jsonschema_spec.handlers.urllib import UrllibHandler as UrlHandler

__all__ = ["FileHandler", "UrlHandler"]

file_handler = FileHandler()
all_urls_handler = UrllibHandler("http", "https", "file")
default_handlers = {
    "<all_urls>": all_urls_handler,
    "http": UrlHandler("http"),
    "https": UrlHandler("https"),
    "file": UrllibHandler("file"),
}



"""JSONSchema spec accessors module."""
from collections import deque
from contextlib import contextmanager
from typing import Any
from typing import Deque
from typing import Hashable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Union

from jsonschema.validators import RefResolver
from pathable.accessors import LookupAccessor

from jsonschema_spec.utils import is_ref


class SpecAccessor(LookupAccessor):
    def __init__(self, lookup: Mapping[Hashable, Any], resolver: RefResolver):
        super().__init__(lookup)
        self.resolver = resolver

    @contextmanager
    def open(
        self, parts: List[Hashable]
    ) -> Iterator[Union[Mapping[Hashable, Any], Any]]:
        parts_deque = deque(parts)
        try:
            yield self._open(self.lookup, parts_deque)
        finally:
            pass

    def _open(
        self, content: Mapping[Hashable, Any], parts_deque: Deque[Hashable]
    ) -> Any:
        if is_ref(content):
            ref = content["$ref"]
            with self.resolver.resolving(ref) as resolved:
                return self._open(resolved, parts_deque)

        try:
            part = parts_deque.popleft()
        except IndexError:
            return content
        else:
            target = content[part]
            return self._open(target, parts_deque)



# Use CSafeFile if available
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from yaml import SafeLoader
else:
    try:
        from yaml import CSafeLoader as SafeLoader
    except ImportError:
        from yaml import SafeLoader


__all__ = [
    "SafeLoader",
]



"""JSONSchema spec handlers requests module."""
from contextlib import closing
from io import StringIO
from typing import Any

import requests

from jsonschema_spec.handlers.file import BaseFilePathHandler


class UrlRequestsHandler(BaseFilePathHandler):
    """URL (requests) scheme handler."""

    def __init__(self, *allowed_schemes: str, **options: Any):
        self.timeout = options.pop("timeout", 10)
        super().__init__(**options)
        self.allowed_schemes = list(allowed_schemes)

    def _open(self, url: str) -> Any:
        response = requests.get(url, timeout=self.timeout)
        response.raise_for_status()

        data = StringIO(response.text)
        with closing(data) as fh:
            return super().__call__(fh)



"""JSONSchema spec paths module."""
from typing import Any
from typing import Hashable
from typing import Mapping

from jsonschema.validators import RefResolver
from pathable.paths import AccessorPath

from jsonschema_spec.accessors import SpecAccessor
from jsonschema_spec.handlers import default_handlers
from jsonschema_spec.handlers.protocols import SupportsRead
from jsonschema_spec.readers import FilePathReader
from jsonschema_spec.readers import FileReader

SPEC_SEPARATOR = "#"


class Spec(AccessorPath):
    @classmethod
    def from_dict(
        cls,
        data: Mapping[Hashable, Any],
        *args: Any,
        spec_url: str = "",
        ref_resolver_handlers: Mapping[str, Any] = default_handlers,
        separator: str = SPEC_SEPARATOR,
    ) -> "Spec":
        ref_resolver = RefResolver(
            spec_url, data, handlers=ref_resolver_handlers
        )
        accessor = SpecAccessor(data, ref_resolver)
        return cls(accessor, *args, separator=separator)

    @classmethod
    def from_file_path(cls, file_path: str) -> "Spec":
        reader = FilePathReader(file_path)
        data, spec_url = reader.read()
        return cls.from_dict(data, spec_url=spec_url)

    @classmethod
    def from_file(cls, fileobj: SupportsRead, spec_url: str = "") -> "Spec":
        reader = FileReader(fileobj)
        data, _ = reader.read()
        return cls.from_dict(data, spec_url=spec_url)



"""JSONSchema spec handlers file module."""
from io import StringIO
from json import dumps
from json import loads
from typing import IO
from typing import Any
from typing import List
from typing import Type
from typing import Union
from urllib.parse import urlparse

from yaml import load

from jsonschema_spec.handlers.compat import SafeLoader
from jsonschema_spec.handlers.protocols import SupportsRead
from jsonschema_spec.handlers.utils import uri_to_path


class FileHandler:
    """File-like object handler."""

    def __init__(self, loader: Type[SafeLoader] = SafeLoader):
        self.loader = loader

    def __call__(self, f: SupportsRead) -> Any:
        data = self._load(f)
        return loads(dumps(data))

    def _load(self, f: SupportsRead) -> Any:
        return load(f, self.loader)


class BaseFilePathHandler(FileHandler):
    """Base file path handler."""

    allowed_schemes: List[str] = NotImplemented

    def __call__(self, uri: Union[SupportsRead, str]) -> Any:
        if isinstance(uri, SupportsRead):
            return super().__call__(uri)

        parsed_url = urlparse(uri)
        if parsed_url.scheme not in self.allowed_schemes:
            raise ValueError("Not allowed scheme")

        return self._open(uri)

    def _open(self, uri: str) -> Any:
        raise NotImplementedError


class FilePathHandler(BaseFilePathHandler):
    """File path handler."""

    allowed_schemes = ["file"]

    def _open(self, uri: str) -> Any:
        filepath = uri_to_path(uri)
        with open(filepath) as fh:
            return super().__call__(fh)



from jsonschema_spec.handlers import default_handlers
from jsonschema_spec.paths import Spec

__author__ = "Artur Maciag"
__email__ = "maciag.artur@gmail.com"
__version__ = "0.1.1"
__url__ = "https://github.com/p1c2u/jsonschema-spec"
__license__ = "Apache License, Version 2.0"

__all__ = ["Spec", "default_handlers"]



"""JSONSchema spec handlers urllib module."""
from typing import Any
from urllib.request import urlopen

from jsonschema_spec.handlers.file import BaseFilePathHandler


class UrllibHandler(BaseFilePathHandler):
    """URL (urllib) scheme handler."""

    def __init__(self, *allowed_schemes: str, **options: Any):
        self.timeout = options.pop("timeout", 10)
        super().__init__(**options)
        self.allowed_schemes = list(allowed_schemes)

    def _open(self, url: str) -> Any:
        with urlopen(url, timeout=self.timeout) as fh:
            return super().__call__(fh)



from typing import Any
from typing import Hashable
from typing import Mapping
from typing import Optional


def is_ref(item: Optional[Mapping[Hashable, Any]]) -> bool:
    return isinstance(item, dict) and "$ref" in item and item["$ref"].__hash__



"""JSONSchema spec readers module."""
from os import path
from pathlib import Path
from typing import Any
from typing import Hashable
from typing import Mapping
from typing import Tuple

from jsonschema_spec.handlers import all_urls_handler
from jsonschema_spec.handlers import file_handler
from jsonschema_spec.handlers.protocols import SupportsRead


class BaseReader:
    def read(self) -> Tuple[Mapping[Hashable, Any], str]:
        raise NotImplementedError


class FileReader(BaseReader):
    def __init__(self, fileobj: SupportsRead):
        self.fileobj = fileobj

    def read(self) -> Tuple[Mapping[Hashable, Any], str]:
        return file_handler(self.fileobj), ""


class FilePathReader(BaseReader):
    def __init__(self, file_path: str):
        self.file_path = file_path

    def read(self) -> Tuple[Mapping[Hashable, Any], str]:
        if not path.isfile(self.file_path):
            raise OSError(f"No such file: {self.file_path}")

        filename = path.abspath(self.file_path)
        uri = Path(filename).as_uri()
        return all_urls_handler(uri), uri



import os.path
import urllib.parse
import urllib.request


def uri_to_path(uri: str) -> str:
    parsed = urllib.parse.urlparse(uri)
    host = "{0}{0}{mnt}{0}".format(os.path.sep, mnt=parsed.netloc)
    return os.path.normpath(
        os.path.join(
            host,
            urllib.request.url2pathname(urllib.parse.unquote(parsed.path)),
        )
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jsonschema_spec', 'jsonschema_spec.handlers']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.1',
 'jsonschema>=4.0.0,<5.0.0',
 'pathable>=0.4.1,<0.5.0',
 'typing-extensions>=4.3.0,<5.0.0']

setup_kwargs = {
    'name': 'jsonschema-spec',
    'version': '0.1.1',
    'description': 'JSONSchema Spec with object-oriented paths',
    'long_description': '***************\nJSONSchema Spec\n***************\n\n.. image:: https://img.shields.io/pypi/v/jsonschema-spec.svg\n     :target: https://pypi.python.org/pypi/jsonschema-spec\n.. image:: https://travis-ci.org/p1c2u/jsonschema-spec.svg?branch=master\n     :target: https://travis-ci.org/p1c2u/jsonschema-spec\n.. image:: https://img.shields.io/codecov/c/github/p1c2u/jsonschema-spec/master.svg?style=flat\n     :target: https://codecov.io/github/p1c2u/jsonschema-spec?branch=master\n.. image:: https://img.shields.io/pypi/pyversions/jsonschema-spec.svg\n     :target: https://pypi.python.org/pypi/jsonschema-spec\n.. image:: https://img.shields.io/pypi/format/jsonschema-spec.svg\n     :target: https://pypi.python.org/pypi/jsonschema-spec\n.. image:: https://img.shields.io/pypi/status/jsonschema-spec.svg\n     :target: https://pypi.python.org/pypi/jsonschema-spec\n\nAbout\n#####\n\nJSONSchema Spec with object-oriented paths\n\nKey features\n************\n\n* Traverse elements like paths\n* Access spec on demand with separate dereferencing accessor layer\n\nInstallation\n############\n\n::\n\n    $ pip install jsonschema-spec\n\nAlternatively you can download the code and install from the repository:\n\n.. code-block:: bash\n\n   $ pip install -e git+https://github.com/p1c2u/jsonschema-spec.git#egg=jsonschema_spec\n\n\nUsage\n#####\n\n.. code-block:: python\n\n   from jsonschema_spec import Spec\n   \n   d = {\n       "openapi": "3.0.1",\n       "info": {\n            "$ref": "#/components/Version",\n       },\n       "paths": {},\n       "components": {\n           "Version": {\n               "title": "Minimal",\n               "version": "1.0",\n            },\n       },\n   }\n   \n   spec = Spec.from_dict(d)\n   \n   # Concatenate paths with /\n   info = spec / "info"\n   \n   # Stat path keys\n   "title" in info\n   \n   # Open path dict\n   with info.open() as info_dict:\n       print(info_dict)\n\n\nRelated projects\n################\n\n* `openapi-core <https://github.com/p1c2u/openapi-core>`__\n   Python library that adds client-side and server-side support for the OpenAPI.\n* `openapi-spec-validator <https://github.com/p1c2u/openapi-spec-validator>`__\n   Python library that validates OpenAPI Specs against the OpenAPI 2.0 (aka Swagger) and OpenAPI 3.0 specification\n* `openapi-schema-validator <https://github.com/p1c2u/openapi-schema-validator>`__\n   Python library that validates schema against the OpenAPI Schema Specification v3.0.\n\nLicense\n#######\n\nCopyright (c) 2017-2022, Artur Maciag, All rights reserved. Apache v2\n',
    'author': 'Artur Maciag',
    'author_email': 'maciag.artur@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/p1c2u/jsonschema-spec',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



from typing import TYPE_CHECKING
from typing import Optional

if TYPE_CHECKING:
    from typing_extensions import Protocol
    from typing_extensions import runtime_checkable
else:
    try:
        from typing import Protocol
        from typing import runtime_checkable
    except ImportError:
        from typing_extensions import Protocol
        from typing_extensions import runtime_checkable


@runtime_checkable
class SupportsRead(Protocol):
    def read(self, amount: Optional[int] = 0) -> str:
        ...
